//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "NetF.h"

inline
void CPartFile::MOD_Destruct_Post()
{
	POSITION pos;
	while ((pos = m_ClientFileStatusList.GetHeadPosition()) != 0)
	{
		CClientFileStatus* fileStatus = m_ClientFileStatusList.GetAt(pos);
		delete fileStatus;
	}
}

inline
void CPartFile::MOD_Init_Post()
{
	m_ics_filemode = 0;	// enkeyDEV: ICS
	// netfinity: Invalidate any eventual gap cache entries
	for (int i = 0; i < m_partStat.GetSize(); i++)
	{
		m_partStat[i].gapCacheValid = false;
		m_partStat[i].hashedBlocks = 0;
	}
}

inline
void CPartFile::MOD_CheckGapCache(uint16 part)
{
	if(part >= m_partStat.GetSize())
	{
		m_partStat.SetSize(part >= GetPartCount() ? part + 1 : max(GetPartCount(), m_partStat.GetSize()));
		for (int i = 0; i < m_partStat.GetSize(); i++)
		{
			m_partStat[i].gapCacheValid = false;
			m_partStat[i].hashedBlocks = 0;
		}
	}
}

inline
void CPartFile::MOD_InvalidateGapCache(uint16 part)
{
	MOD_CheckGapCache(part);
	m_partStat[part].gapCacheValid = false;
}

inline
void CPartFile::MOD_SetGapCache(uint16 part, uint32 size)
{
	MOD_CheckGapCache(part);
	m_partStat[part].gapSize = size;
	m_partStat[part].gapCacheValid = true;
}

inline
void CPartFile::MOD_CreatePartFile_Post()
{
	m_SrcIncPartFrequency.SetSize(GetPartCount()); // enkeyDEV: ICS
	for (uint32 i = 0; i < GetPartCount();i++)
		m_SrcIncPartFrequency.Add(0);	// enkeyDEV: ICS
}

inline
void CPartFile::MOD_LoadPartFile_Post()
{
	m_SrcIncPartFrequency.SetSize(GetPartCount()); // enkeyDEV: ICS
	for (uint32 i = 0; i != GetPartCount();i++)
		m_SrcIncPartFrequency.Add(0);	//enkeyDEV: ICS
}

inline
void CPartFile::MOD_AddGap_Post(uint32 start, uint32 end)
{
	for(uint32 i = (start / PARTSIZE); i <= (end / PARTSIZE); ++i)
		if(start <= i * PARTSIZE && end >= (i + 1) * PARTSIZE - 1)
			MOD_SetGapCache(i,PARTSIZE);
		else
			MOD_InvalidateGapCache(i);
}

inline
void CPartFile::MOD_FillGap_Post(uint32 start, uint32 end)
{
	for(uint32 i = (start / PARTSIZE); i <= (end / PARTSIZE); ++i)
		if(start <= i * PARTSIZE && end >= (i + 1) * PARTSIZE - 1)
			MOD_SetGapCache(i,0);
		else
			MOD_InvalidateGapCache(i);
}

uint32 CPartFile::GetTotalGapSizeInPart(UINT uPart)
{
	uint32 uRangeStart = uPart * PARTSIZE;
	uint32 uRangeEnd = uRangeStart + PARTSIZE - 1;
	if (uRangeEnd >= m_nFileSize)
		uRangeEnd = m_nFileSize;
	// MOD BEGIN netfinity:
	if(uRangeStart > uRangeEnd) return 0;
	MOD_CheckGapCache(uPart);
	if (m_partStat[uPart].gapCacheValid == true)
	{
		return m_partStat[uPart].gapSize;
	}
	m_partStat[uPart].gapCacheValid = true;
	return (m_partStat[uPart].gapSize = GetTotalGapSizeInRange(uRangeStart, uRangeEnd));
	// MOD END netfinity
}

bool CPartFile::GetNextEmptyBlockInPart(uint16 partNumber, Requested_Block_Struct *result, uint32 bytesRequested, uint32 lastOffset) /*const*/
{
	//try
	//{
	//Gap_Struct *firstGap = NULL;
	Gap_Struct *currentGap;
	uint32 start;
	uint32 end;
	
	// Find start of this part
	uint32 partStart = (PARTSIZE * partNumber);
	// What is the end limit of this block, i.e. can't go outside part (or filesize)
	uint32 partEnd = (PARTSIZE * (partNumber + 1)) - 1;
	if (partEnd >= GetFileSize())
		partEnd = GetFileSize() - 1;

	uint32 lastBlock = (lastOffset >= partStart && lastOffset <= partEnd ? (lastOffset - partStart) / EMBLOCKSIZE : ~0);

	if(partStart > partEnd)
	{
		AddDebugLogLine(false, _T("%s; Part beyond end of file!"),_T(__FUNCTION__));
		return false;
	}

	// Don't bother to check complete parts
	if(GetTotalGapSizeInPart(partNumber) == 0)
	{
		// Part is complete
		return false;
	}

	// Fast return if no block offset is needed
	if(result == NULL)
	{
		if (MOD_GetRequestedSizeInPart(partNumber) < GetTotalGapSizeInPart(partNumber))
			return true; // There are still unrequested blocks
		else
			return false; // All blocks have been requested
	}

	CList<struct dataBlock>	availableBlocks;

	// Find the first gap from the start position
	for (POSITION pos = gaplist.GetHeadPosition(); pos != 0; )
	{
		currentGap = gaplist.GetNext(pos);
		if(currentGap == NULL)
			break; // We are done!
		if(currentGap->start <= partEnd && currentGap->end >= partStart)
		{
			for(uint32 i = 0; i < 53; ++i)
			{
				uint32	blockStart = partStart + i * EMBLOCKSIZE;
				uint32	blockEnd = min(blockStart + EMBLOCKSIZE - 1, partEnd);
				if(blockStart > partEnd) break;
				if ((currentGap->start <= blockEnd) && (currentGap->end >= blockStart))
				{
					uint32				start = max(blockStart, currentGap->start);
					uint32				end = min(blockEnd, currentGap->end);
					struct dataBlock	newBlock;
					newBlock.start = start;
					newBlock.end = end;
					availableBlocks.AddTail(newBlock);
				}
			}
		}
	}

	for (POSITION i =  requestedblocks_list.GetHeadPosition();i != 0; )
	{
		const Requested_Block_Struct* cur_block = requestedblocks_list.GetNext(i);
		for (POSITION j =  availableBlocks.GetHeadPosition();j != 0; )
		{
			const dataBlock avail_block = availableBlocks.GetAt(j);
			if (avail_block.start <= cur_block->EndOffset && avail_block.end >=cur_block->StartOffset)
			{
				availableBlocks.RemoveAt(j);
				struct dataBlock	newBlock;
				if (avail_block.start < cur_block->StartOffset)
				{
					newBlock.start = avail_block.start;
					newBlock.end = cur_block->StartOffset - 1;
					availableBlocks.AddTail(newBlock);
				}
				if (avail_block.end > cur_block->EndOffset)
				{
					newBlock.start = cur_block->EndOffset + 1;
					newBlock.end = avail_block.end;
					availableBlocks.AddTail(newBlock);
				}
				j =  availableBlocks.GetHeadPosition();
			}
			else
				availableBlocks.GetNext(j);
		}
	}

	if(availableBlocks.IsEmpty())
	{
		AddDebugLogLine(false, _T("%s; No block found"),_T(__FUNCTION__));
		AddDebugLogLine(false, _T("%s; Part requested: %d  Part left: %d"),_T(__FUNCTION__), MOD_GetRequestedSizeInPart(partNumber), GetTotalGapSizeInPart(partNumber));
		return false;
	}

	uint32 block = (rand() + ::GetTickCount()) % availableBlocks.GetCount(); // netfinity: Use time to get more randomness
	POSITION idx = availableBlocks.FindIndex(block);
	const dataBlock avail_block = availableBlocks.GetAt(idx);
	start = avail_block.start;
	end = avail_block.end;

	if (lastBlock != ~0)
	{
		for (idx =  availableBlocks.GetHeadPosition(); idx != 0; )
		{
			const dataBlock avail_block = availableBlocks.GetNext(idx);
			if((avail_block.start - partStart) / EMBLOCKSIZE == lastBlock)
			{
				start = avail_block.start;
				end = avail_block.end;
				break;
			}
		}
	}

	bytesRequested -= bytesRequested % 10240; 
	if (bytesRequested < 10240) bytesRequested = 10240;
	if (bytesRequested > EMBLOCKSIZE) bytesRequested = EMBLOCKSIZE;
	if (start + bytesRequested - 1 < end)
		end = start + bytesRequested - 1;

	if(start > end || (start - partStart) / EMBLOCKSIZE != (end - partStart) / EMBLOCKSIZE)
	{
		AddDebugLogLine(false, _T("%s; Block boundaries voilated: start = %d, stop = %d"),_T(__FUNCTION__), start, end);
		return false;
	}

	//AddDebugLogLine(false, _T("Found block = %d"), i);
	if (result != NULL)
	{
		result->StartOffset = start;
		result->EndOffset = end;
		md4cpy(result->FileID, GetFileHash());
		result->transferred = 0;
	}
	return true;
	//}
	//catch(...)
	//{
	//	AddDebugLogLine(false, _T("CPartFile::GetNextEmptyBlockInPart - Unknown exception"));
	//	return false;
	//}
}

// netfinity: Cleaned up the CPartFile::GetNextEmptyBlockInPart()
//  a little bit and put some of the code here 
Gap_Struct* CPartFile::MOD_FindGap(uint32 start, uint32 end) const
{
	Gap_Struct *firstGap = NULL;
	Gap_Struct *currentGap;

	// Find the first gap from the start position
	for (POSITION pos = gaplist.GetHeadPosition(); pos != 0; )
	{
		currentGap = gaplist.GetNext(pos);
		// Want gaps that overlap start<->partEnd
		if ((currentGap->start <= end) && (currentGap->end >= start))
		{
			// Is this the first gap?
			if ((firstGap == NULL) || (currentGap->start < firstGap->start))
			{
				firstGap = currentGap;
				break; // We are done!
			}
		}
	}
	return firstGap;
}

// enkeyDEV: ICS
void CPartFile::MOD_WriteIncPartStatus(CSafeMemFile* file){
	uint16 parts = GetPartCount();
	file->WriteUInt16(parts);
	uint16 done = 0;
	while (done != parts){
		uint8 towrite = 0;
		for (uint32 i = 0;i != 8;i++){
			if (!IsPureGap(done*PARTSIZE,((done+1)*PARTSIZE)-1))
				towrite |= (1<<i);
			done++;
			if (done == parts)
				break;
		}
		file->WriteUInt8(towrite);
	}
}
// <--- enkeyDEV: ICS

// MOD BEGIN netfinity: Download bandwidth throttler
/**
 * @brief Processes downloads to test for timeouts and to calculate download rate
 *
 * Note! Only to be called from CPartFile::Process()
 */
void CPartFile::MOD_ProcessDL(CUpDownClient *cur_src, uint32 reducedownload, CBytesPerSecond allowed)
{
	CBytesPerSecond	cur_datarate, new_datarate;

	ASSERT(cur_src->socket);
	if (cur_src->socket)
	{
		cur_src->CheckDownloadTimeout();
		cur_datarate = cur_src->CalculateDownloadRate();
		datarate+=cur_datarate;
		if(reducedownload)
		{
			float	reduce_modifier = ((float) reducedownload - 100.0f) / 100.0f;
			// Boost rare parts
			float	part_modifier = 1.0f;
			if(cur_src->m_lastPartAsked != 0xFFFF)
			{
				uint16	sourceCount = m_SrcpartFrequency[cur_src->m_lastPartAsked];
				if(sourceCount <= 5)
					part_modifier = 1.30f;
				else if (sourceCount <= 10)
					part_modifier = 1.20f;
				else if (sourceCount <= 30)
					part_modifier = 1.10f;
				else if (sourceCount <= 100)
					part_modifier = 1.05f;
				allowed = allowed * (part_modifier / 1.30f);
				float	completion = ((float) PARTSIZE - GetTotalGapSizeInPart(cur_src->m_lastPartAsked)) / ((float) PARTSIZE);
				part_modifier *= (0.2 * completion + 1.0);
				if(reduce_modifier < 0.0f) part_modifier = 1.0f / part_modifier;
			}
			// Calculate new data rate
			new_datarate = (reduce_modifier * part_modifier + 1.0f) * cur_datarate;
			if(new_datarate < allowed)
				new_datarate = allowed;	

			//AddDebugLogLine(false, _T("new_datarate = %d"), new_datarate);

			cur_src->socket->MOD_SetDownloadLimitByRate(new_datarate);
			if (cur_src->IsDownloadingFromPeerCache() && cur_src->m_pPCDownSocket && cur_src->m_pPCDownSocket->IsConnected())
				cur_src->m_pPCDownSocket->MOD_SetDownloadLimitByRate(new_datarate);
		}
		else{
			cur_src->socket->DisableDownloadLimit();
			if (cur_src->IsDownloadingFromPeerCache() && cur_src->m_pPCDownSocket && cur_src->m_pPCDownSocket->IsConnected())
				cur_src->m_pPCDownSocket->DisableDownloadLimit();
		}
	}
}
// MOD END netfinity

// MOD BEGIN netfinity: Read out source count
uint16 CPartFile::MOD_GetPartSourceCount(uint16 part)
{
	if(!IsPartFile()) return CKnownFile::MOD_GetPartSourceCount(part);
	if(part >= m_SrcpartFrequency.GetCount()) return 0;
	return m_SrcpartFrequency[part];
}
// MOD END netfinity

// enkeyDEV: ICS - Intelligent Chunk Selection
void CPartFile::MOD_NewSrcIncPartsInfo()
{
	uint16 partcount = GetPartCount();

	if (m_SrcIncPartFrequency.GetSize() < partcount)
		m_SrcIncPartFrequency.SetSize(partcount);

	for(int i = 0; i < partcount; i++)
		m_SrcIncPartFrequency[i] = 0;

	CClientFileStatus* pFileStatus;

	for (POSITION pos = m_ClientFileStatusList.GetHeadPosition(); pos != NULL;){
		pFileStatus = m_ClientFileStatusList.GetNext(pos);
		for (int i = 0; i < partcount; i++){
			if (pFileStatus->IsIncompletePart(i))
				m_SrcIncPartFrequency[i] += 1;
		}
	}
	//UpdateDisplayedInfo(); // Not displayed
}

#define	CM_RELEASE_MODE			1
#define	CM_SPREAD_MODE			2
#define	CM_SHARE_MODE			3

#define	CM_SPREAD_MINSRC		10
#define	CM_SHARE_MINSRC			25
#define CM_MAX_SRC_CHUNK		3

uint16 CPartFile::MOD_FindNextPart(CUpDownClient* sender)
{
	// Select mode: RELEASE, SPREAD or SHARE

	uint16	part_idx;
	uint16	min_src = 0xFFFF;

	if (m_SrcpartFrequency.GetCount() < GetPartCount())
		min_src = 0;
	else
		for (part_idx = 0; part_idx < GetPartCount(); ++part_idx)
			if (m_SrcpartFrequency[part_idx] < min_src)
				min_src = m_SrcpartFrequency[part_idx];

	if (min_src <= CM_SPREAD_MINSRC)		m_ics_filemode = CM_RELEASE_MODE;
	else if (min_src <= CM_SHARE_MINSRC)	m_ics_filemode = CM_SPREAD_MODE;
	else									m_ics_filemode = CM_SHARE_MODE;

	// Chunk list ordered by preference

	//CList<uint16,uint16> chunk_list;
	//CList<uint32,uint32> chunk_pref;
	double c_pref;
	uint32 complete_src;
	uint32 incomplete_src;
	uint32 first_last_mod;
	uint32 size2transfer;
	uint32 downloading_src;
	uint32 downloading_file_src;
	//bool already_enough_requests = false;

	uint16 best_chunk = 0xFFFF;
	double best_score = 1000000000.0;

	downloading_file_src = MOD_GetDownloadingSourcesInFile();

	// MOD BEGIN netfinity: Step up one level for part sources
	if (sender->MOD_IsPartialSource())
	{
		if (m_ics_filemode == CM_SPREAD_MODE) m_ics_filemode = CM_SHARE_MODE;
		if (m_ics_filemode == CM_RELEASE_MODE) m_ics_filemode = CM_SPREAD_MODE;
	}
	// MOD END netfinity

	const bool isPreviewEnable = (thePrefs.GetPreviewPrio() || thePrefs.IsExtControlsEnabled() && GetPreviewPrio()) && IsPreviewableFileType(); // netfinity: Added from 0.44b

	for (part_idx = 0; part_idx < GetPartCount(); ++part_idx)
	{
		// netfinity: Anti HideOS
		if ((theNetF.m_bDownloadHidden && sender->MOD_IsPartialSource() ?
			 sender->m_pDownFileStatus->IsSeenPart(part_idx) : 
			 sender->m_pDownFileStatus->IsCompletePart(part_idx)
			) &&
			GetNextEmptyBlockInPart(part_idx, 0)
		   )
		{
			complete_src = 0;
			incomplete_src = 0;
			first_last_mod = 0;

			// Chunk priority modifiers
			if (m_ics_filemode != CM_SHARE_MODE)
			{
				complete_src = m_SrcpartFrequency.GetCount() > part_idx ? m_SrcpartFrequency[part_idx] : 0;
				incomplete_src = m_SrcIncPartFrequency.GetCount() > part_idx ? m_SrcIncPartFrequency[part_idx] : 0;
			}
			if (m_ics_filemode != CM_RELEASE_MODE)
			{
				// netfinity: this seemed completly flawed!
				if (part_idx == 0 || part_idx == (GetPartCount() - 1))		first_last_mod = 0;
				else if (part_idx == 1 || part_idx == (GetPartCount() - 2))	first_last_mod = 1;
				else														first_last_mod = 2;
				if (!isPreviewEnable) first_last_mod = 0;
			}
			size2transfer = GetTotalGapSizeInPart(part_idx);
			// netfinity: Don't assign all sources to same part
			downloading_src = MOD_GetDownloadingSourcesInPart(part_idx, sender);
			
			// Add up modifiers
			//size2transfer = min(((size2transfer + 0xff) >> 8), 0xFFFF);
			double fact_left = size2transfer / PARTSIZE; 
			switch (m_ics_filemode)
			{
			case CM_RELEASE_MODE:
				c_pref = (complete_src + incomplete_src * 0.5) * 10.0 + fact_left;
				if(downloading_src > downloading_file_src / 2 + 1)
					c_pref += 1000000.0;
				break;
			case CM_SPREAD_MODE:
				if(downloading_src <= downloading_file_src / 4)
					c_pref = (complete_src + incomplete_src * 0.5) * 10.0 + fact_left * 5.0 + first_last_mod;
				else
					c_pref = 1000000.0 + (complete_src + incomplete_src * 0.5) * 10.0 + 5.0 - fact_left * 5.0 + first_last_mod;
				break;
			case CM_SHARE_MODE:
				if(downloading_src <= downloading_file_src / 8)
					c_pref = fact_left * 5.0 + first_last_mod;
				else
					c_pref = 1000000.0 + 5.0 - fact_left * 5.0 + first_last_mod;
				break;
			}
			if (size2transfer < 3 * EMBLOCKSIZE)
				c_pref += 10000000.0;
			c_pref += (double)rand()/(double)RAND_MAX; // netfinity: Avoid having to chunks with same preference

			if (c_pref < best_score)
			{
				best_score = c_pref;
				best_chunk = part_idx;
			}
		}
	}

	return best_chunk;
}

bool CPartFile::GetNextRequestedBlock(CUpDownClient* sender, Requested_Block_Struct** newblocks, uint16* count, uint32 pendingBytes)
{
	if (!(*count)) return false;					// added as in 29a
//	if (!(sender->GetPartStatus())) return false;	// added as in 29a

	bool	freshClient = (*count == 3);
	uint16	requestedCount = *count;
	uint16	newblockcount = 0;
	*count = 0;

	// MOD BEGIN netfinity: Seed random generator
	if (sender->m_lastPartAsked == 0xFFFF)
		srand((unsigned int) sender + ::GetTickCount());
	// MOD END netfinity

	// MOD BEGIN netfinity: Check if an other part is to be chosen
	if (sender->m_lastPartAsked == 0xFFFF || MOD_GetRequestedSizeInPart(sender->m_lastPartAsked) >= GetTotalGapSizeInPart(sender->m_lastPartAsked))
	{
		if ((sender->m_lastPartAsked = MOD_FindNextPart(sender)) == 0xFFFF) 
		{
			//AddDebugLogLine(false, _T("PartFile::GetNextRequestedBlock - No needed part"));
			return false;
		}
		//AddDebugLogLine(false, _T("PartFile::GetNextRequestedBlock - New part chosen: part = %d"), sender->m_lastPartAsked);
		if (sender->m_lastPartAsked == 0xFFFF)
			AddDebugLogLine(false, _T("%s (FindNextPart); sender->m_lastPartAsked == 0xFFFF"), _T(__FUNCTION__));
	}
	// MOD END netfinity

	// MOD BEGIN netfinity: Dynamic Block Requests - Calculate block request size
	uint32	bytesToRequest = 3 * EMBLOCKSIZE;
	uint32	bytesToDownloadInPart = GetTotalGapSizeInPart(sender->m_lastPartAsked) - MOD_GetRequestedSizeInPart(sender->m_lastPartAsked);
	uint32	bytesToDownloadInFile = GetTotalGapSizeInRange(0, GetFileSize()) - MOD_GetRequestedSizeInFile();
	uint16	downloadingSourcesInPartExcl = MOD_GetDownloadingSourcesInPart(sender->m_lastPartAsked, sender);
	uint32	sourceDatarate = sender->GetDownloadDatarate().GetValue();
	uint32	partDatarateExcl = MOD_GetDownloadingDatarateInPart(sender->m_lastPartAsked, sender);
	//uint32	fileDatarate = GetDatarate();
	float	rateFactor = 1.0;
	if (partDatarateExcl > 0)
		rateFactor = (float) sourceDatarate / (float) (partDatarateExcl + sourceDatarate);
	bytesToRequest = min(bytesToRequest, bytesToDownloadInPart * rateFactor / 2);
	bytesToRequest = min(bytesToRequest, sourceDatarate * 60); 
	bytesToRequest = min(bytesToRequest, bytesToDownloadInPart / (downloadingSourcesInPartExcl + 2));
	if (bytesToDownloadInPart < 6 * EMBLOCKSIZE)
	{
		if (sourceDatarate < 1024 || bytesToRequest < 10240)
			bytesToRequest = 10240;
	}
	else if (bytesToDownloadInFile < 3 * PARTSIZE)
	{
		if (sourceDatarate < 1024 || bytesToRequest < 3 * 10240)
			bytesToRequest = 3 * 10240;
	}
	else
	{
		if (sourceDatarate < 1024 || bytesToRequest < EMBLOCKSIZE)
			bytesToRequest = EMBLOCKSIZE;
	}
	if (bytesToRequest <= pendingBytes)
	{
		if (requestedCount < 2)
		{
			//AddDebugLogLine(false, _T("PartFile::GetNextRequestedBlock - Request already large enough"));
			return false;
		}
		else
			bytesToRequest = 10240;
	}
	else
		bytesToRequest -= pendingBytes;
	if (freshClient && sourceDatarate < 1024 && bytesToRequest <= EMBLOCKSIZE) // Don't split initial requests. Ensures better compability with Edonkey's
		requestedCount = 1;
	// Check if we should drop download to speed up file completion
	if (downloadingSourcesInPartExcl > 0 && sourceDatarate < 1024 && partDatarateExcl > 1024)
	{
		if (sourceDatarate < 200) sourceDatarate = 200;
		uint32	timeToPartCompletionExcl = bytesToDownloadInPart / partDatarateExcl;
		uint32	timeToFileCompletionExcl = bytesToDownloadInFile / partDatarateExcl; // not exact but only interresting when at the very last bytes
		uint32	timeToBlockCompletion = 10240 / sourceDatarate;
		if (!freshClient && timeToBlockCompletion > timeToFileCompletionExcl + 10)
		{
			AddDebugLogLine(false, _T("%s; Too slow"), _T(__FUNCTION__));
			sender->m_lastPartAsked = 0xffff;
			return false;
		}
		else if (freshClient && timeToFileCompletionExcl < 20)
		{
			AddDebugLogLine(false, _T("%s; Too close to file completion"), _T(__FUNCTION__));
			sender->m_lastPartAsked = 0xffff;
			return false;
		}
	}
	// MOD END netfinity

	if (sender->m_lastPartAsked == 0xFFFF)
		AddDebugLogLine(false, _T("%s; sender->m_lastPartAsked == 0xFFFF"),_T(__FUNCTION__));
	if (newblockcount >= requestedCount)
		AddDebugLogLine(false, _T("%s; newblockcount >= requestedCount"),_T(__FUNCTION__));
	if (bytesToRequest == 0)
		AddDebugLogLine(false, _T("%s; bytesToRequest == 0"),_T(__FUNCTION__));

	// MOD BEGIN netfinity: Remember last asked part (from Maella)
	while(sender->m_lastPartAsked != 0xFFFF && newblockcount < requestedCount && bytesToRequest > 0)
	{
		Requested_Block_Struct* pBlock = new Requested_Block_Struct;
		if(GetNextEmptyBlockInPart(sender->m_lastPartAsked, pBlock, bytesToRequest / (requestedCount - newblockcount), sender->m_lastBlockOffset) == true)
		{
			// netfinity: Decrease bytes to request
			uint32	bytesRequested = pBlock->EndOffset - pBlock->StartOffset + 1;
			if (bytesToRequest >= bytesRequested)
				bytesToRequest -= bytesRequested;
			else
				bytesToRequest = 0;
			// Keep track of last requested block offset
			sender->m_lastBlockOffset = pBlock->StartOffset;
			// Keep a track of all pending requested blocks
			requestedblocks_list.AddTail(pBlock);
			// Update list of blocks to return
			newblocks[newblockcount++] = pBlock;
			*count = newblockcount;
			if (newblockcount == requestedCount)
				return true;
			// Skip end of loop (=> CPU load)
			continue;
		} 
		else
		{
			// All blocks for this chunk have been already requested
			delete pBlock;
			// => Try to select another chunk
			AddDebugLogLine(false, _T("%s; No block returned"),_T(__FUNCTION__));
			sender->m_lastPartAsked = 0xffff;
			//if(!MOD_FindNextPart(sender))
			//	break;
		}
	}
	// MOD END netfinity

	if (!(*count))
	{
		AddDebugLogLine(false, _T("%s; Unexpected NNP"),_T(__FUNCTION__));
		return false;
	}
	return true;
}

uint32 CPartFile::MOD_GetRequestedSizeInPart(uint16 part) const
{
	uint32	requestedSize = 0;
	// Find start of this part
	uint32 partStart = (PARTSIZE * part);
	// What is the end limit of this block, i.e. can't go outside part (or filesize)
	uint32 partEnd = (PARTSIZE * (part + 1)) - 1;
	if (partEnd >= GetFileSize())
		partEnd = GetFileSize() - 1;

	for (POSITION pos =  requestedblocks_list.GetHeadPosition();pos != 0; )
	{
		const Requested_Block_Struct* cur_block = requestedblocks_list.GetNext(pos);
		if ((partStart <= cur_block->StartOffset) && (partEnd >= cur_block->EndOffset))
		{
			requestedSize += (cur_block->EndOffset - cur_block->StartOffset + 1) - cur_block->transferred;
		}
	}
	return requestedSize;
}

uint32 CPartFile::MOD_GetRequestedSizeInFile() const
{
	uint32	requestedSize = 0;

	for (POSITION pos =  requestedblocks_list.GetHeadPosition();pos != 0; )
	{
		const Requested_Block_Struct* cur_block = requestedblocks_list.GetNext(pos);
		requestedSize += (cur_block->EndOffset - cur_block->StartOffset + 1) - cur_block->transferred;
	}
	return requestedSize;
}

uint16 CPartFile::MOD_GetDownloadingSourcesInPart(uint16 part, CUpDownClient* exclude_src) const
{
	uint16	sourceCount = 0;
	for(POSITION pos = m_downloadingSourceList.GetHeadPosition();pos!=0;)
	{
		CUpDownClient* cur_src = m_downloadingSourceList.GetNext(pos);
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID( cur_src );
		if (cur_src && cur_src->GetDownloadState() == DS_DOWNLOADING && cur_src != exclude_src)
		{
			if (cur_src->m_lastPartAsked == part)
				sourceCount++;
		}
	}
	return sourceCount;
}

uint16 CPartFile::MOD_GetDownloadingSourcesInFile(CUpDownClient* exclude_src) const
{
	uint16	sourceCount = 0;
	for(POSITION pos = m_downloadingSourceList.GetHeadPosition();pos!=0;)
	{
		CUpDownClient* cur_src = m_downloadingSourceList.GetNext(pos);
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID( cur_src );
		if (cur_src && cur_src->GetDownloadState() == DS_DOWNLOADING && cur_src != exclude_src)
		{
			sourceCount++;
		}
	}
	return sourceCount;
}

uint32 CPartFile::MOD_GetDownloadingDatarateInPart(uint16 part, CUpDownClient* exclude_src) const
{
	uint32	dataRate = 0;
	for(POSITION pos = m_downloadingSourceList.GetHeadPosition();pos!=0;)
	{
		CUpDownClient* cur_src = m_downloadingSourceList.GetNext(pos);
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID( cur_src );
		if (cur_src && cur_src->GetDownloadState() == DS_DOWNLOADING && cur_src != exclude_src)
		{
			if (cur_src->m_lastPartAsked == part)
				dataRate += cur_src->GetDownloadDatarate().GetValue();
		}
	}
	return dataRate;
}

bool CPartFile::MOD_DropStalledDownloads(CUpDownClient* pClient) const
{
	bool	dropped = false;
	uint32	fileDatarate = GetDatarate().GetValue();
	uint32	fileDownloading = MOD_GetDownloadingSourcesInFile();
	uint32	avgDatarate = 0;
	uint32	srcDatarate = 0;
	if (fileDownloading != 0)
		avgDatarate = fileDatarate / fileDownloading;
	if (pClient->GetDownloadDatarate() < 1024)
		return false;
	CClientFileStatus* pStatus = pClient->m_FileStatusList.FindFile(this);
	if (pStatus == NULL)
		return false;
	for(POSITION pos = m_downloadingSourceList.GetHeadPosition();pos!=0;)
	{
		CUpDownClient* cur_src = m_downloadingSourceList.GetNext(pos);
		if (thePrefs.m_iDbgHeap >= 2)
			ASSERT_VALID( cur_src );
		if (cur_src && cur_src != pClient && cur_src->GetDownloadState() == DS_DOWNLOADING)
		{
			if ( cur_src->m_dwDownStartTime.Elapsed() > SEC2MS(10)) // 10 seconds of download has passed
			{
				srcDatarate = cur_src->GetDownloadDatarate().GetValue();
				if (!cur_src->m_dwNoNeededPartsPending.IsTickSet() // This source is not in NNP pending state
					&& pStatus->IsSeenPart(cur_src->m_lastPartAsked) // And we can download this part (TODO! have to implement a better method here)
					&& (cur_src->GetTransferredDownMini() == 0  // And nothing was received
						|| (srcDatarate < 1024 && srcDatarate < 2 * avgDatarate && fileDatarate > 5 * srcDatarate) // Or it's really slow
						|| cur_src->socket->timeout_timer.Elapsed() > SEC2MS(5) // Or more than 5 seconds elapsed since last activity on socket 
					   )
				   )
				{
					AddDebugLogLine(false, _T(__FUNCTION__) _T("; Download dropped"));
					cur_src->SendCancelTransfer();
					cur_src->SetDownloadState(DS_ONQUEUE, REASON_Dropped);
					dropped = true;
				}
			}
		}
	}
	return true;
}

void CPartFile::MOD_OnPartHashValid(uint16 partNumber, bool bForceICH, bool bNoAICH)
{
	uint32 partRange = (m_hpartfile.GetLength() < ((uint32) partNumber) * PARTSIZE ) ? (m_hpartfile.GetLength() % PARTSIZE) - 1 : (PARTSIZE - 1);
	MOD_CheckGapCache(partNumber); // netfinity: Check we have a part record for this part
	// Is this 9MB part complete
	if (GetTotalGapSizeInPart(partNumber) == 0) //IsComplete(PARTSIZE * partNumber, (PARTSIZE * (partNumber + 1)) - 1))
	{
		if (!hashsetneeded){
			if (thePrefs.GetVerbose())
				AddDebugLogLine(DLP_VERYLOW, false, _T("Finished part %u of \"%s\""), partNumber, GetFileName());
		}

		// tell the blackbox about the verified data
		m_CorruptionBlackBox.VerifiedData(PARTSIZE*partNumber, PARTSIZE*partNumber + partRange);

		// if this part was successfully completed (although ICH is active), remove from corrupted list
		POSITION posCorrupted = corrupted_list.Find(partNumber);
		if (posCorrupted)
			corrupted_list.RemoveAt(posCorrupted);

		m_partStat[partNumber].hashedBlocks = ~0; // netfinity: Tag part as verified
		if (status == PS_EMPTY)
		{
			if (theApp.emuledlg->IsRunning()) // may be called during shutdown!
			{
				if (GetHashCount() == GetED2KPartHashCount() && !hashsetneeded)
			{
					// Successfully completed part, make it available for sharing
					SetStatus(PS_READY);
					theApp.sharedfiles->SafeAddKFile(this);
				}
			}
		}
	}
	else if (IsCorruptedPart(partNumber) && (thePrefs.IsICHEnabled() || bForceICH))
	{
		// Try to recover with minimal loss
		m_uPartsSavedDueICH++;
		thePrefs.Add2SessionPartsSavedByICH(1);

		uint32 uRecovered = GetTotalGapSizeInPart(partNumber);
		FillGap(PARTSIZE*partNumber, PARTSIZE*partNumber + partRange);
		RemoveBlockFromList(PARTSIZE*partNumber, PARTSIZE*partNumber + partRange);

		// tell the blackbox about the verified data
		m_CorruptionBlackBox.VerifiedData(PARTSIZE*partNumber, PARTSIZE*partNumber + partRange);

		// remove from corrupted list
		POSITION posCorrupted = corrupted_list.Find(partNumber);
		if (posCorrupted)
			corrupted_list.RemoveAt(posCorrupted);

		AddLogLine(true, GetResString(IDS_ICHWORKED), partNumber, GetFileName(), CastItoXBytes(uRecovered, false, false));

		// correct file stats
		if (m_uCorruptionLoss >= uRecovered) // check, in case the tag was not present in part.met
			m_uCorruptionLoss -= uRecovered;
		// here we can't know if we have to subtract the amount of recovered data from the session stats
		// or the cumulative stats, so we subtract from where we can which leads eventuall to correct 
		// total stats
		if (thePrefs.sesLostFromCorruption >= uRecovered)
			thePrefs.sesLostFromCorruption -= uRecovered;
		else if (thePrefs.cumLostFromCorruption >= uRecovered)
			thePrefs.cumLostFromCorruption -= uRecovered;

		m_partStat[partNumber].hashedBlocks = ~0; // netfinity: Tag part as verified
		if (status == PS_EMPTY)
		{
			if (theApp.emuledlg->IsRunning()) // may be called during shutdown!
			{
				if (GetHashCount() == GetED2KPartHashCount() && !hashsetneeded)
				{
					// Successfully recovered part, make it available for sharing
					SetStatus(PS_READY);
					theApp.sharedfiles->SafeAddKFile(this);
				}
			}
		}
	}
}

void CPartFile::MOD_OnPartHashInvalid(uint16 partNumber, bool bForceICH, bool bNoAICH)
{
	uint32 partRange = (m_hpartfile.GetLength() < ((uint32) partNumber) * PARTSIZE ) ? (m_hpartfile.GetLength() % PARTSIZE) - 1 : (PARTSIZE - 1);
	MOD_CheckGapCache(partNumber); // netfinity: Check we have a part record for this part
	// Is this 9MB part complete
	if (GetTotalGapSizeInPart(partNumber) == 0) //IsComplete(PARTSIZE * partNumber, (PARTSIZE * (partNumber + 1)) - 1))
	{
		LogWarning(LOG_STATUSBAR, GetResString(IDS_ERR_PARTCORRUPT), partNumber, GetFileName());
		AddGap(PARTSIZE*partNumber, PARTSIZE*partNumber + partRange);

		// add part to corrupted list, if not already there
		if (!IsCorruptedPart(partNumber))
			corrupted_list.AddTail(partNumber);

		// request AICH recovery data
		if (!bNoAICH)
			RequestAICHRecovery((uint16)partNumber);

		// update stats
		m_uCorruptionLoss += (partRange + 1);
		thePrefs.Add2LostFromCorruption(partRange + 1);
	}
}